#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Desc: { 项目验证类模块 }

import logging

from lmgrun import lmgwin
from lmgrun.execute import execute
from lmgrun.baseinfo import BaseInfo
from lmgrun.feedback import feedback
from lmgrun.lmgenum import StatusCodeEnum
from lmgrun.exceptions import LoongkerImageNotFound

log = logging.getLogger(__name__)

class Validator(BaseInfo):
    """ Validates plan for various errors like availability of loongker images etc.
    """
    def __init__(self):
        self.img_name = BaseInfo.img_name
        self.cntr_name = BaseInfo.cntr_name

    def validate(self):
        # Check loongker installed
        self.check_can_run_loongker()
        self.check_loongker_image_exists()
        code = self.check_loongker_container_exists()
        return code

    def check_can_run_loongker(self):
        """
        Checks if lmgrun can run loongker.
        """
        cmd, retcode, output = '<none>', 0, ['<none>']
        try:
            cmd = ["loongker", "--version"]
            result = execute(cmd)
            output = result.out
            retcode = result.return_code
            log.debug('lmgrun can run loongker:%s %d %s' %(cmd, retcode, output))
        except OSError as error:
            data = {
                'code': StatusCodeEnum.CHECK_RUN_DOCKER_ERR.code,
                'errmsg': StatusCodeEnum.CHECK_RUN_DOCKER_ERR.errmsg
            }
            msg = 'CanRunLoongker:{} {} {} {} {}' .format(data['code'], cmd, retcode, output, str(error))
            log.error(msg)
            prompt_msg = 'Error code:{} {}' .format(data['code'], data['errmsg'])
            lmgwin.lmg_prompt('error', prompt_msg)
            #feedback()
            raise

    def check_loongker_image_exists(self):
        """
        Before cold start, you need to check whether the loongker images exists
        """
        runcmd = ["loongker",  "inspect",  "--type=image",  self.img_name]
        try:
            result = execute(runcmd)
            if not result.succeeded :
                raise LoongkerImageNotFound(self.img_name)
            return result.succeeded
        except OSError as error:
            data = {
                'code': StatusCodeEnum.CHECK_IMAGE_CMD_ERR.code,
                'errmsg': StatusCodeEnum.CHECK_IMAGE_CMD_ERR.errmsg
            }
            msg = "LoongkerImageExists:{} (image={}) " "check failed with message: '{}'" .format(data['code'], self.img_name, str(error))
            log.error(msg)
            prompt_msg = 'Error code:{} {}' .format(data['code'], data['errmsg'])
            lmgwin.lmg_prompt('error', prompt_msg)
            #feedback()
            raise

    def check_loongker_container_exists(self):
        """
        Before cold start, you need to check whether the loongker images exists
        """
        runcmd = ["loongker",  "inspect",  "--type=container",  self.cntr_name]
        try:
            result = execute(runcmd)
            if not result.succeeded:
                log.debug("LoongkerContainerNotExists (container=%s) " % (self.cntr_name))
            return result.succeeded
        except OSError as error:
            data = {
                'code': StatusCodeEnum.CHECK_CONTAINER_CMD_ERR.code,
                'errmsg': StatusCodeEnum.CHECK_CONTAINER_CMD_ERR.errmsg
            }
            msg = "LoongkerContainereExists:{} (container={}) " "check failed with message: '{}'" .format(data['code'], self.cntr_name, str(error))
            log.error(msg)
            prompt_msg = 'Error code:{} {}' .format(data['code'], data['errmsg'])
            lmgwin.lmg_prompt('error', prompt_msg)
            #feedback()
            raise
