import docker
import time
import logging
log = logging.getLogger(__name__)

from appexceptions import appNotFound, appNotRunning

class AppContainer(object):
    """
    Class for accessing information about a app's container.
    """
    def __init__(self, iname, cname, docker_url="unix://var/run/docker.sock"):
        self.iname = iname
        self.cname = cname
        self.docker_url = docker_url

    def get_docker_client(self):
        try:
            log.debug("Connecting to docker daemon on local unix socket")
            client = docker.DockerClient(base_url=self.docker_url, version='auto')
        except docker.errors.APIError as e:
            log.debug("Could not connect to docker daemon:" + e)
            raise appException(
                "On Demand backend unavailible. Please contact an admin."
            )
        return client

    def get_api_client(self):
        try:
            log.debug("Connecting to docker daemon on local unix socket")
            client = docker.APIClient(base_url=self.docker_url, version='auto')
        except docker.errors.APIError as e:
            log.debug("Could not connect to docker daemon:" + e)
            raise appException(
                "On Demand backend unavailible. Please contact an admin."
            )
        return client

    def get_id(self):
        """
        Look up the container ID as used by Docker.
        """
        info = self.inspect()

        return info['Id']

    def get_ip(self):
        """
        Look up the IP address assigned to the container.
        """
        info = self.inspect()

        if not info['State']['Running']:
            raise appNotRunning("The app is not running.")

        return info['NetworkSettings']['IPAddress']

    def get_pid(self):
        """
        Look up the PID of the container, if running.
        """
        info = self.inspect()

        if not info['State']['Running']:
            raise appNotRunning("The app is not running.")

        return info['State']['Pid']

    def get_status(self):
        """
        Return the status of the container (running, exited, paused).

        Returns "missing" if the app does not exist.
        """
        try:
            info = self.inspect()
            return info['State']['Status']
        except appNotFound:
            return "missing"

    def inspect(self):
        """
        Return the full container status from Docker.
        """
        client = self.get_api_client()
        try:
            info = client.inspect_container(self.cname)
            return info
        except docker.errors.NotFound:
            raise appNotFound("The app could not be found.")

    def is_running(self):
        """
        Check if container is running.

        Returns True/False; returns False if the container does not exist.
        """
        try:
            info = self.inspect()
            return info['State']['Running']
        except appNotFound:
            return False

    def app_container_start(self, detach, command, environment, devices, volumes,  kwargs):
        client = self.get_docker_client()
        try:
            container = client.containers.run(detach=detach, command=command, image=self.iname, name=self.cname, environment=environment, devices=devices, volumes=volumes, **kwargs)
            log.info("Successfully started container with Id: %s\n" % (str(container.id)))
        except Exception as e:
            raise e

    def app_container_rm(self):
        client = self.get_docker_client()
        try:
            container = client.containers.get(self.cname)
            container.remove(force=True)
        except Exception as error:
            log.debug("Error removing container: {}".format(error))
