import os
import re
import sys
import pwd
import shutil
import fnmatch
import traceback
import platform
import logging
import requests
import configparser
from tqdm import tqdm
from configobj import ConfigObj
from .appc import AppImage, load_image
from ..inst.lmgformat import Lmgformat, UnPackage
from ..utils.execute import execute, get_lpkg_path, get_depend_path
from .lmgfile import Lmgstatus, Appindex, Applist, Lmgmeta, Repofile
from ..utils.exceptions import DockerLoadError, LpkgError, PackError,  DependError, AppRunError

log = logging.getLogger(__name__)

def get_desktop_applications():
    return '/usr/share/applications'

def get_desktop_default():
    return '/opt/lmg'

def get_applist_dir():
    return '/var/lib/lpkg/info'

def get_applist_file(appname):
    return '/var/lib/lpkg/info/' + appname +  '.list'

def get_dependence_dir():
    return './'

def get_icon_dir():
    return '/usr/share/icons/lmg'

def get_appindex_file():
    return '/var/lib/lpkg/appindex.cfg'

def get_lmgstatus_file():
    return '/var/lib/lpkg/status'


def lpkg_init():
    if not os.path.isdir(get_applist_dir()):
        os.makedirs(get_applist_dir())

    if not os.path.isdir(get_icon_dir()):
        os.makedirs(get_icon_dir())

    if not os.path.isdir(get_desktop_default()):
        os.makedirs(get_desktop_default())

    if not os.path.isdir(os.path.join(get_desk_path(), 'lmg')):
        os.symlink(get_desktop_default(), 
             os.path.join(get_desk_path(), 'lmg'))
        os.chown(os.path.join(get_desk_path(), 'lmg'),
            pwd.getpwnam(os.getlogin()).pw_uid,
            pwd.getpwnam(os.getlogin()).pw_gid)
        
    if not os.path.isdir(get_lpkg_path()):
        os.makedirs(get_lpkg_path())

    if not os.path.isdir(get_depend_path()):
        os.makedirs(get_depend_path())

    if not os.path.isfile(get_appindex_file()):
        os.mknod(get_appindex_file(), 0o644)

    if not os.path.isfile(get_lmgstatus_file()):
        os.mknod(get_lmgstatus_file(), 0o644)

def get_desk_path():
    home = pwd.getpwnam(os.getlogin()).pw_dir
    dirconf = os.path.join(home, '.config/user-dirs.dirs')
    config = ConfigObj(dirconf)
    homedesk = config.get('XDG_DESKTOP_DIR')

    if os.path.isdir(home + '/Desktop') and not homedesk:
        homedesk = home + '/Desktop'
    elif not homedesk:
        homedesk = home + '/桌面'
    else:
        homedesk = os.path.join(
                home, 
                os.path.basename(homedesk))
    return homedesk

class Lmgman(object):
    def __init__(self, Lmgstatus, Appindex, Applist = None):
        self.lmgstatus = Lmgstatus
        self.appindex = Appindex
        self.applist = Applist

    def list_app(self):
        self.lmgstatus.yaml_print()

    def list_app_file(self, appname):
        if self.applist:
            self.appindex.yaml_print('appname', appname)
            self.applist.yaml_print()


class LmgInst(Lmgman):
    def __init__(self, lmgstatus, appindex, applist, lmgmeta, lmgformat, sourcepath, destpath):
        super(LmgInst, self).__init__(lmgstatus, appindex, applist)
        self.sourcepath = sourcepath
        self.destpath = destpath
        self.lmgformat = lmgformat

        self.applayer = lmgmeta.get_value('applayer')
        self.appname = lmgmeta.get_value('appname')
        self.appversion = lmgmeta.get_value('appversion')
        self.arch = lmgmeta.get_value('arch')
        self.apptag = lmgmeta.get_value('apptag')
        self.lmgtype = lmgmeta.get_value('lmgtype')
        self.dependence = lmgmeta.get_value('dependence')
        self.description = lmgmeta.get_value('description')
        self.filelist = {'apptag':[], 'files':[], 'dirs':[]}


    def __check_arch(self):
        os_arch = platform.machine()
        #if os_arch == 'loongarch64':
        #    os_arch = 'loong64'

        if self.arch != os_arch:
            raise PackError('Arch mismatch !')

    def __check_sign(self):
        pass

    def __check_file(self):
        pass

    def check_lmg(self):
        self.__check_arch()
        self.__check_sign()
        self.__check_file()
    
    def repo_search(self, url, fname):
        resp = requests.get(url, stream = True, timeout = 5)
        if resp.status_code == requests.codes.ok:
            with open(fname, 'wb') as file:
                for data in resp.iter_content(chunk_size=1024):
                    file.write(data)
            repofile = Repofile(fname)
            return repofile.yaml_search(self.dependence, self.arch)
        else:
            resp.raise_for_status()

    def down_file(self, url, fname):
        resp = requests.get(url, stream = True, timeout = 5)
        if resp.status_code == requests.codes.ok:
            total = int (resp.headers.get('content-length', 0))
            with open(fname, 'wb') as file, tqdm(
                    desc=fname, 
                    total=total,
                    unit='iB',
                    unit_scale=True,
                    unit_divisor=1024,
                    ) as bar:
                for data in resp.iter_content(chunk_size=1024):
                    size = file.write(data)
                    bar.update(size)
            return True
        else:
            resp.raise_for_status()
            return False

    def base_from_repo(self):
        #repourl = 'http://10.2.5.28/os/lmg/baseruntime.yaml'
        repourl = 'http://lcf.loongnix.cn/os/lmg/baseruntime.yaml'
        downurl = 'http://lcf.loongnix.cn/os/lmg/baseruntime/%s/%s_brl.%s.lmg' %(self.arch, self.dependence, self.arch)
        if self.repo_search(repourl, 'repofile'):
            return self.down_file(downurl, '%s_brl.%s.lmg' %(self.dependence, self.arch))

#TODO: if statement solve
    def process_depend(self):
        if  self.dependence and not self.lmgstatus.yaml_isexist('app', self.dependence):
            log.info('Processing the ' + self.appname + ' dependencies ' + self.dependence)
            depend_pathname = os.path.join(get_dependence_dir(),
                                '%s_brl.%s.lmg' %(self.dependence, self.arch))

            if (os.path.isfile(depend_pathname) or 
               self.base_from_repo()):
                lmgfactory = LmgFactory()
                lmgoption = LmgOption(lmgfactory)
                lmgoption.CreateLmgInst(depend_pathname, udir = os.path.join(get_lpkg_path(), 'tmpone'))
                
                if  self.dependence and not self.lmgstatus.yaml_isexist('app', self.dependence):
                    raise DependError('The %s needs to install %s dependencies, but the installation failed!' %(self.appname, self.dependence))
            else:
                raise DependError('The %s needs to install %s dependencies !' %(self.appname, self.dependence))

    def inst_image(self):
        log.info('Start processing ' + self.appname + ' container images')

        image_pathname = self.lmgformat.get_file_path('image')
        if  image_pathname:
            load_image(os.path.join(self.sourcepath, image_pathname),
                    self.lmgtype)

            if self.lmgtype == 'BASERUNTIME':
                self.filelist['apptag'].append(self.apptag)
                #self.appindex.yaml_up_apptag()

            elif self.lmgtype == 'APPLICATIONS':
                appimage = AppImage(self.applayer)
                self.filelist['apptag'] = appimage.set_image_name()[:]

            elif self.lmgtype == 'UPBASERUNTIME':
                self.filelist['apptag'].append(self.apptag)
                self.appindex.yaml_up_apptag()

            else:
                raise PackError('lmgtype error: %s ' %self.lmgtype)

            os.remove(self.sourcepath + '/' + image_pathname)

    def _clear_exec(self, string):
        return re.sub(r' -- |^.*sh -c |\"', '', string).strip().strip('%').replace('\'', '\"')

    def change_desk(self, deskfile, icon_pathname_half):
        config = configparser.RawConfigParser(interpolation=None, strict=False)
        config.optionxform = lambda option: option
        config.read(deskfile, encoding='UTF8')
        #get and prepare value
        dicon = os.path.join(get_icon_dir(), icon_pathname_half)
        for section in config.sections():
            if 'Path' in config[section].keys():
                command = 'cd {};{}'.format(config.get(section, 'Path'),
                        self._clear_exec(config[section]['Exec']))
            else:
                command = self._clear_exec(config[section]['Exec'])

            dexec = 'bash -c \'/usr/local/bin/lmgrun -p {name}_{version} -c ' \
                    'bash -l -c \"{command}\"\''.format(name=self.appname.replace('%', '%%'),
                            version=self.appversion.replace('%', '%%'),
                            command=command) 
            #set value
            config.remove_option(section,'Path')
            config.remove_option(section,'TryExec')
            config.set(section, 'Icon', dicon)
            config.set(section, 'Exec', dexec)
        with open(deskfile, 'w') as f:
            config.write(f)


    def inst_desk(self):
        desktops = self.lmgformat.get_desktop_dict()
        if not os.path.isdir(get_desktop_default()):
            os.makedirs(get_desktop_default())

        if not os.path.isdir(os.path.join(get_desk_path(), 'lmg')):
            os.symlink(get_desktop_default(), 
                     os.path.join(get_desk_path(), 'lmg'))
            os.chown(os.path.join(get_desk_path(), 'lmg'),
                    pwd.getpwnam(os.getlogin()).pw_uid,
                    pwd.getpwnam(os.getlogin()).pw_gid)

        for desk,icon in desktops.items():
            desk_pathname_new = os.path.join(get_desktop_default(), os.path.basename(desk).rsplit('.', 1)[0] + '_' + self.appversion + '.desktop')
            icon_pathname_half = os.path.basename(icon).rsplit('.', 1)[0] + '_' + self.appversion  + os.path.splitext(icon)[1]
            self.change_desk(os.path.join(self.sourcepath, desk),icon_pathname_half)

            shutil.move('%s' %(self.sourcepath + '/' + desk), desk_pathname_new)
            os.chmod(desk_pathname_new,  0o755)
            os.chown(desk_pathname_new,
                    pwd.getpwnam(os.getlogin()).pw_uid,
                    pwd.getpwnam(os.getlogin()).pw_gid)
            self.filelist['files'].append(desk_pathname_new)

            # two desktop -----> one icon
            if os.path.isfile(os.path.join(self.sourcepath, icon)):
                os.chmod(os.path.join(self.sourcepath, icon), 0o644)
                shutil.move(os.path.join(self.sourcepath, icon),
                            get_icon_dir()+ '/' + icon_pathname_half)
                self.filelist['files'].append(os.path.join(get_icon_dir(), icon_pathname_half))


    def inst_file(self):
        self.inst_image()
        self.inst_desk()
        if os.path.isdir(self.destpath + '/' + 
                os.path.basename(self.lmgformat.path)):

            shutil.rmtree(self.destpath + '/' + 
                    os.path.basename(self.lmgformat.path))
        shutil.move(self.lmgformat.path, self.destpath)
        
        self.filelist['dirs'].append(self.destpath + '/' + os.path.basename(self.lmgformat.path))

    def record_info(self):

        log.info('Record the ' + self.appname + ' information')
        lmgstatus_document = [{'app':self.appname + '_' + self.appversion, 'applayer':self.applayer, 'appname': self.appname,
        'appversion': self.appversion, 'arch': self.arch, 'dependence': self.dependence, 'description': self.description}]
        appindex_document =[{'app':self.appname + '_' + self.appversion, 'appname':self.appname, 'applayer':self.applayer, 'meta':self.destpath + '/' + self.lmgformat.get_file_path('meta'),
            'apptag':self.filelist['apptag'], 'lmgtype':self.lmgtype}]
        self.lmgstatus.yaml_add(lmgstatus_document)
        self.appindex.yaml_add(appindex_document)
        self.applist.yaml_dump(self.filelist)

    def inst_app(self):
        if not self.lmgstatus.yaml_isexist('app', self.appname + '_' + self.appversion):
            # log.info('Start installing the software')
            self.check_lmg()
            self.process_depend()
            self.inst_file()
            self.record_info()
            log.info('The ' + self.appname + ' installation is complete!\n')
        else:
            log.warning('The ' + self.appname + ' already exists!')

class Lmguninst(Lmgman):
    def __init__(self, lmgstatus, appindex, applist, appname):
        super(Lmguninst, self).__init__(lmgstatus, appindex, applist)
        self.appname = appname

    def uninst_app(self):
        if not self.lmgstatus.yaml_isexist('appname', self.appname):
            log.warning("The %s is not installed !" %self.appname)
        #elif self.lmgstatus.yaml_isexist('dependence', self.appname+'_'+self.appversion):
        #    log.warning("The %s is relied on !" %self.appname)
        else:
            log.info('Start removing the %s container images' %self.appname)
            self.appindex.yaml_dele('appname', self.appname)

            log.info('Delete the %s installation records' %self.appname)
            self.applist.rmfile()
            self.lmgstatus.yaml_dele('appname', self.appname)

            log.info("The %s uninstalled successfully !" %self.appname)

class LmgFactory(object):
    def MakeUnPackage(self, pathname, udir):
        return UnPackage(pathname, udir)

    def MakeLmgformat(self, sourcepath):
        return Lmgformat(sourcepath)

    def MakeAppindex(self, yamlfile):
        return Appindex(yamlfile)

    def MakeLmgstatus(self, yamlfile):
        return Lmgstatus(yamlfile)

    def MakeApplist(self, yamlfile):
        return Applist(yamlfile)

    def MakeLmgmeta(self, yamlfile):
        return Lmgmeta(yamlfile)

    def MakeLmgman(self, Lmgstatus, Appindex, Applist = None):
        return Lmgman(Lmgstatus, Appindex, Applist)

    def MakeLmgInst(self, Lmgstatus, Appindex, Applist, Lmgmeta, Lmgformat, sourcepath, destpath):
        return LmgInst(Lmgstatus, Appindex, Applist, Lmgmeta, Lmgformat, sourcepath, destpath)

    def MakeLmgUnInst(self, Lmgstatus, Appindex, Applist, appname):
        return Lmguninst(Lmgstatus, Appindex, Applist, appname)


class LmgOption(object):
    def __init__(self, LmgFactory):
        self.lmgfactory = LmgFactory

    def CreateLmgInst(self, pathname, destpath = '/opt',
                      udir = os.path.join(get_lpkg_path(), 'tmp')):
        sourcepath = udir
        try:
            log.info('Start installing the software.')
            lmgstatus = self.lmgfactory.MakeLmgstatus(get_lmgstatus_file())
            appindex = self.lmgfactory.MakeAppindex(get_appindex_file())

            self.lmgfactory.MakeUnPackage(pathname, udir).do_unpack()
            lmgformat = self.lmgfactory.MakeLmgformat(sourcepath)
            lmgformat.load_dict()
            lmgformat.load_desktop()
            lmgformat.check_file_exist()

            yamlfile = os.path.join(sourcepath, lmgformat.get_file_path('meta'))
            lmgmeta = self.lmgfactory.MakeLmgmeta(yamlfile)
            lmgmeta.yaml_load()
            yamlfile = get_applist_file(lmgmeta.get_value('appname'))
            applist = self.lmgfactory.MakeApplist(yamlfile)

            lmg = self.lmgfactory.MakeLmgInst(lmgstatus, appindex, applist, lmgmeta, lmgformat, sourcepath, destpath)
            lmg.inst_app()
            return 0

        except Exception as e:
            #log.exception(e)
            log.debug(traceback.format_exc())
            log.error(repr(e))
            log.error('The ' + pathname + ' installation failed !' )
            sys.exit(1)

        finally:
            if os.path.isdir(sourcepath):
                shutil.rmtree(sourcepath)

    def CreateLmgUnInst(self, appname):
        try:
            lmgstatus = self.lmgfactory.MakeLmgstatus(get_lmgstatus_file())
            appindex = self.lmgfactory.MakeAppindex(get_appindex_file())
            applist = self.lmgfactory.MakeApplist(get_applist_file(appname))
            lmg = self.lmgfactory.MakeLmgUnInst(lmgstatus, appindex, applist, appname)
            lmg.uninst_app()
            return 0

        except AppRunError:
            os.remove('%s.bak' %get_appindex_file())
            log.error('The %s is running.' %appname)
            sys.exit(1)

        except Exception as e:
            log.debug(traceback.format_exc())
            log.error(repr(e))
            log.error('The %s uninstall error.\nPlease use -f to try force uninstall!!!' %appname)
            sys.exit(1)

    def CreateLmgList(self):
        try:
            lmgstatus = self.lmgfactory.MakeLmgstatus(get_lmgstatus_file())
            appindex = self.lmgfactory.MakeAppindex(get_appindex_file())
            lmg = self.lmgfactory.MakeLmgman(lmgstatus, appindex)
            lmg.list_app()
            return 0

        except Exception as e:
            # log.exception(e)
            log.debug(traceback.format_exc())
            log.error(repr(e))
            log.error(e)
            sys.exit(1)

    def CreateLmgListFile(self, appname):
        try:
            lmgstatus = self.lmgfactory.MakeLmgstatus(get_lmgstatus_file())
            appindex = self.lmgfactory.MakeAppindex(get_appindex_file())
            applist = self.lmgfactory.MakeApplist(get_applist_file(appname))
            lmg = self.lmgfactory.MakeLmgman(lmgstatus, appindex, applist)
            lmg.list_app_file(appname)
            return 0

        except Exception as e:
            log.debug(traceback.format_exc())
            log.error(repr(e))
            log.error(e)
            sys.exit(1)

    def CreateLmgField(self, appname):
        try:
            yamlfile = None
            udir = os.path.join(get_lpkg_path(), 'tmptwo')
            self.lmgfactory.MakeUnPackage(appname, udir).do_field()
            for filename in os.listdir(udir):
                if filename.endswith('.app'):
                    directory = os.path.join(udir, filename)
                    metaname = filename.rsplit('.', 1)[0] + '.yaml'
                    yamlfile = os.path.join(directory, metaname)
                    break
            if yamlfile:
                lmgmeta = self.lmgfactory.MakeLmgmeta(yamlfile)
                lmgmeta.yaml_load()
                lmgmeta.yaml_print()
            else:
                raise LpkgError('Lmgmeta file missing')

        except Exception as e:
            log.debug(traceback.format_exc())
            log.error(repr(e))
            log.error(e)
            sys.exit(1)

        finally:
            shutil.rmtree(udir)
